"""Implementation for DisplayObject class.
"""

from apysc._animation.animation_parallel_mixin import AnimationParallelMixIn
from apysc._display.css_interface import CssInterface
from apysc._display.parent_mixin import ParentMixIn
from apysc._display.visible_mixin import VisibleMixIn
from apysc._event.custom_event_mixin import CustomEventMixIn
from apysc._event.mouse_event_mixins import MouseEventMixIns
from apysc._html.debug_mode import add_debug_info_setting
from apysc._validation import arg_validation_decos


class DisplayObject(
    ParentMixIn,
    MouseEventMixIns,
    VisibleMixIn,
    CustomEventMixIn,
    CssInterface,
    AnimationParallelMixIn,
):
    """
    Display object (base) class for the common interfaces.

    References
    ----------
    - DisplayObject
        - https://simon-ritchie.github.io/apysc/en/display_object.html
    """

    @arg_validation_decos.not_empty_string(arg_position_index=1)
    @add_debug_info_setting(module_name=__name__)
    def __init__(self, *, variable_name: str) -> None:
        """
        Display object (base) class for the common interfaces.

        Parameters
        ----------
        variable_name : str
            Variable name of this instance. A js expression
            uses this setting.

        References
        ----------
        - DisplayObject
            - https://simon-ritchie.github.io/apysc/en/display_object.html
        """
        self._variable_name = variable_name



"""Class implementation for the SVG text's text mix-in.
"""

from typing import Dict
from typing import Optional

from typing_extensions import final

from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.revert_mixin import RevertMixIn
from apysc._type.string import String
from apysc._type.variable_name_mixin import VariableNameMixIn
from apysc._validation import arg_validation_decos


class SvgTextTextMixIn(
    VariableNameMixIn,
    RevertMixIn,
):
    _text: str = ""

    @property
    @add_debug_info_setting(module_name=__name__)
    def text(self) -> String:
        """
        Get a current text's string.

        Returns
        -------
        text : String
            A current text's string.
        """
        from apysc._type.variable_name_suffix_utils import (
            get_attr_or_variable_name_suffix,
        )

        suffix: str = get_attr_or_variable_name_suffix(
            instance=self,
            value_identifier="text",
        )
        text: String = String(self._text, variable_name_suffix=suffix)
        self._append_text_getter_expression(text=text)
        return text

    @text.setter
    @arg_validation_decos.is_apysc_string(arg_position_index=1)
    @add_debug_info_setting(module_name=__name__)
    def text(self, value: String) -> None:
        """
        Set a current text's string.

        Parameters
        ----------
        value : String
            A text to set.
        """
        self._text = value._value
        self._append_text_setter_expression(text=value)

    @final
    @add_debug_info_setting(module_name=__name__)
    def _append_text_getter_expression(self, *, text: String) -> None:
        """
        Append a text's getter expression string.

        Parameters
        ----------
        text : String
            A target text string.
        """
        from apysc._expression import expression_data_util

        expression: str = f"{text.variable_name} = {self.variable_name}.text();"
        expression_data_util.append_js_expression(expression=expression)

    @final
    @add_debug_info_setting(module_name=__name__)
    def _append_text_setter_expression(self, *, text: String) -> None:
        """
        Append a text's setter expression string.

        Parameters
        ----------
        text : String
            A target text string.
        """
        from apysc._expression import expression_data_util

        expression: str = f"{self.variable_name}.text({text.variable_name});"
        expression_data_util.append_js_expression(expression=expression)

    _text_snapshots: Optional[Dict[str, str]] = None

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._set_single_snapshot_val_to_dict(
            dict_name="_text_snapshots",
            value=self._text,
            snapshot_name=snapshot_name,
        )

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if a snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._text = self._get_snapshot_val_if_exists(
            current_value=self._text,
            snapshot_dict=self._text_snapshots,
            snapshot_name=snapshot_name,
        )



"""The interface implementation for the `_initialize_with_base_value` method.
"""

from abc import abstractmethod
from typing import Any


class InitializeWithBaseValueInterface:
    @classmethod
    @abstractmethod
    def _initialize_with_base_value(cls) -> Any:
        """
        Initialize this class with a base value(s).
        """
        raise NotImplementedError(
            "The _initialize_with_base_value abstract method must be implemented."
        )



"""The class implementation for the `contact_phone` material icon.
"""

from apysc._display.fixed_html_svg_icon_base import FixedHtmlSvgIconBase


class MaterialContactPhoneIcon(FixedHtmlSvgIconBase):
    """
    The class implementation for the `contact_phone` material icon.
    """

    def _get_fixed_svg_icon_html(self) -> str:
        """
        Get a fixed SVG icon HTML string.

        Returns
        -------
        fixed_svg_icon_html : str
            Fixed SVG icon HTML string.

        References
        ----------
        - Material icon document
            - https://simon-ritchie.github.io/apysc/en/material_icon.html
        - Material Symbols & Icons, APACHE LICENSE, VERSION 2.0
            - https://fonts.google.com/icons?icon.size=24&icon.color=%23e8eaed
            - https://www.apache.org/licenses/LICENSE-2.0.html
        """
        return '<svg xmlns="http://www.w3.org/2000/svg" height="24" viewBox="0 0 24 24" width="24"><path d="M0 0h24v24H0z" fill="none"/><path d="M22 3H2C.9 3 0 3.9 0 5v14c0 1.1.9 2 2 2h20c1.1 0 1.99-.9 1.99-2L24 5c0-1.1-.9-2-2-2zM8 6c1.66 0 3 1.34 3 3s-1.34 3-3 3-3-1.34-3-3 1.34-3 3-3zm6 12H2v-1c0-2 4-3.1 6-3.1s6 1.1 6 3.1v1zm3.85-4h1.64L21 16l-1.99 1.99c-1.31-.98-2.28-2.38-2.73-3.99-.18-.64-.28-1.31-.28-2s.1-1.36.28-2c.45-1.62 1.42-3.01 2.73-3.99L21 8l-1.51 2h-1.64c-.22.63-.35 1.3-.35 2s.13 1.37.35 2z"/></svg>'  # noqa



"""This module is for the translation mapping data of the
following document:

Document file: array_slice.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {
    ##################################################
    "# Array class slice interface": "# Array クラスの slice インターフェイス",
    ##################################################
    "This page explains the `Array` class `slice` method interface.": "このページでは`Array`クラスの`slice`メソッドのインターフェイスについて説明します。",  # noqa
    ##################################################
    "## What interface is this?": "## インターフェイス概要",
    ##################################################
    "The `slice` method interface extracts the specified index range array's values and returns a new array.": "`slice`メソッドのインターフェイスは指定されたインデックスの範囲の配列の値を抽出し新しい配列を返却します。",  # noqa
    ##################################################
    "## Basic usage": "## 基本的な使い方",
    ##################################################
    "The `slice` method requires the `start` and `end` arguments (`int` or `Int` values) and returns a new array.": "`slice`メソッドは`start`と`end`の各引数（Pythonビルトインの`int`もしくはapyscの`Int`の整数）を必要とし、返却値として新しい配列を返します。",  # noqa
    ##################################################
    "If you specify 1 to the `start` argument and 3 to the `end` argument, this method behaves like the Python built-in list slice of `[1:3]`.": "例として`start`引数に1を指定し`end`引数に3を指定した場合、このメソッドはPythonビルトインの`[1:3]`という指定によるスライスと同じように動作します。",  # noqa
    ##################################################
    "An original array is not modified.": "元々の配列の値は変更されません。",
    ##################################################
    "```py\n# runnable\nimport apysc as ap\n\nap.Stage()\narr: ap.Array[int] = ap.Array([1, 2, 3, 4])\nsliced_arr: ap.Array[int] = arr.slice(start=1, end=3)\nassert sliced_arr == [2, 3]\nassert arr == [1, 2, 3, 4]\n```": "```py\n# runnable\nimport apysc as ap\n\nap.Stage()\narr: ap.Array[int] = ap.Array([1, 2, 3, 4])\nsliced_arr: ap.Array[int] = arr.slice(start=1, end=3)\nassert sliced_arr == [2, 3]\nassert arr == [1, 2, 3, 4]\n```",  # noqa
    ##################################################
    "## slice API": "## slice API",
    ##################################################
    '<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>': '<span class="inconspicuous-txt">特記事項: このAPIドキュメントはドキュメントビルド用のスクリプトによって自動で生成・同期されています。そのためもしかしたらこの節の内容は前節までの内容と重複している場合があります。</span>',  # noqa
    ##################################################
    "**[Interface summary]**": "**[インターフェイス概要]**",
    ##################################################
    "Slice this array by specified start and end indexes.<hr>": "指定された開始と終了のインデックス範囲でこの配列をスライスします。<hr>",  # noqa
    ##################################################
    "**[Parameters]**": "**[引数]**",
    ##################################################
    "- `start`: Int or int or None, default None": "- `start`: Int or int or None, default None",  # noqa
    ##################################################
    "  - Slicing start index.": "  - スライス範囲の開始インデックス。",
    ##################################################
    "- `end`: Int or int or None, default None": "- `end`: Int or int or None, default None",  # noqa
    ##################################################
    "  - Slicing end index (a result array does not contain this index).": "  - スライス範囲の終了インデックス（結果の配列のこのインデックスの値を含みません）。",  # noqa
    ##################################################
    "<hr>": "<hr>",
    ##################################################
    "**[Returns]**": "**[返却値]**",
    ##################################################
    "- `sliced_arr`: Array": "- `sliced_arr`: Array",
    ##################################################
    "  - Sliced array.": "  - スライスされた配列。",
    ##################################################
    "<hr>": "<hr>",
    ##################################################
    "**[Examples]**": "**[コードサンプル]**",
    ##################################################
    "```py\n>>> import apysc as ap\n>>> _ = ap.Stage()\n>>> arr: ap.Array = ap.Array([1, 2, 3, 4])\n>>> arr.slice(start=1, end=3)\nArray([2, 3])\n\n>>> arr.slice(start=1)\nArray([2, 3, 4])\n\n>>> arr.slice(end=2)\nArray([1, 2])\n```": "```py\n>>> import apysc as ap\n>>> _ = ap.Stage()\n>>> arr: ap.Array = ap.Array([1, 2, 3, 4])\n>>> arr.slice(start=1, end=3)\nArray([2, 3])\n\n>>> arr.slice(start=1)\nArray([2, 3, 4])\n\n>>> arr.slice(end=2)\nArray([1, 2])\n```",  # noqa
}



"""The mix-in class implementation for the axis `line_color` value.
"""

from typing_extensions import final

from apysc._color.color import Color
from apysc._html.debug_mode import add_debug_info_setting
from apysc._validation import arg_validation_decos


class AxisLineColorMixIn:
    @final
    @arg_validation_decos.is_color(arg_position_index=1, optional=False)
    @add_debug_info_setting(module_name=__name__)
    def _set_initial_line_color(
        self,
        *,
        line_color: Color,
    ) -> None:
        """
        Set an initial line color.

        Parameters
        ----------
        line_color : Color
            An axis line color setting.
        """
        self._line_color = line_color.copy()



"""The class implementation for the outlined `print_disabled` material icon.
"""

from apysc._display.fixed_html_svg_icon_base import FixedHtmlSvgIconBase


class MaterialPrintDisabledOutlinedIcon(FixedHtmlSvgIconBase):
    """
    The class implementation for the outlined `print_disabled` material icon.
    """

    def _get_fixed_svg_icon_html(self) -> str:
        """
        Get a fixed SVG icon HTML string.

        Returns
        -------
        fixed_svg_icon_html : str
            Fixed SVG icon HTML string.

        References
        ----------
        - Material icon document
            - https://simon-ritchie.github.io/apysc/en/material_icon.html
        - Material Symbols & Icons, APACHE LICENSE, VERSION 2.0
            - https://fonts.google.com/icons?icon.size=24&icon.color=%23e8eaed
            - https://www.apache.org/licenses/LICENSE-2.0.html
        """
        return '<svg xmlns="http://www.w3.org/2000/svg" height="24" viewBox="0 0 24 24" width="24"><path d="M0 0h24v24H0z" fill="none"/><path d="M1.41 1.6L0 3.01 5 8c-1.66 0-3 1.34-3 3v6h4v4h12l2.95 2.96 1.41-1.41L1.41 1.6zM6 15H4v-4c0-.55.45-1 1-1h2l3 3H6v2zm2 4v-4h4l4 4H8zM8 5h8v3h-5.34l2 2H19c.55 0 1 .45 1 1v4l-2 .01V13h-2.34l4 4H22v-6c0-1.66-1.34-3-3-3h-1V3H6v.36l2 2V5z"/><circle cx="18" cy="11.51" r="1"/></svg>'  # noqa



"""The utilities implementations for the variable name suffix-related
functions.
"""

from typing import Any


def get_attr_or_variable_name_suffix(*, instance: Any, value_identifier: str) -> str:
    """
    Get an attribute or variable name's suffix if a specified instance's
    type is the `VariableNameSuffixAttrOrVarMixIn` and its suffix value
    is not blank.

    Parameters
    ----------
    instance : Any
        A target instance.
    value_identifier : str
        A value identifier string (e.g., `x`).

    Returns
    -------
    attr_or_variable_name_suffix : str
        An attribute or variable's name suffix.
        In the following cases, this value becomes a blank string.
        - If a specified instance is not the `VariableNameSuffixMixIn`
            and `VariableNameSuffixAttrOrVarMixIn` instance.
        - If a suffix is a blank string.
    """
    from apysc._type.variable_name_suffix_attr_or_var_mixin import (
        VariableNameSuffixAttrOrVarMixIn,
    )

    suffix: str = ""
    if isinstance(instance, VariableNameSuffixAttrOrVarMixIn):
        suffix = instance._get_attr_or_variable_name_suffix(
            value_identifier=value_identifier
        )
    return suffix



"""The class implementation for the `plagiarism` material icon.
"""

from apysc._display.fixed_html_svg_icon_base import FixedHtmlSvgIconBase


class MaterialPlagiarismIcon(FixedHtmlSvgIconBase):
    """
    The class implementation for the `plagiarism` material icon.
    """

    def _get_fixed_svg_icon_html(self) -> str:
        """
        Get a fixed SVG icon HTML string.

        Returns
        -------
        fixed_svg_icon_html : str
            Fixed SVG icon HTML string.

        References
        ----------
        - Material icon document
            - https://simon-ritchie.github.io/apysc/en/material_icon.html
        - Material Symbols & Icons, APACHE LICENSE, VERSION 2.0
            - https://fonts.google.com/icons?icon.size=24&icon.color=%23e8eaed
            - https://www.apache.org/licenses/LICENSE-2.0.html
        """
        return '<svg xmlns="http://www.w3.org/2000/svg" enable-background="new 0 0 24 24" height="24" viewBox="0 0 24 24" width="24"><g><rect fill="none" height="24" width="24"/></g><g><g><path d="M14,2H6C4.9,2,4,2.9,4,4v16c0,1.1,0.89,2,1.99,2H18c1.1,0,2-0.9,2-2V8L14,2z M15.04,19.45l-1.88-1.88 c-1.33,0.71-3.01,0.53-4.13-0.59c-1.37-1.37-1.37-3.58,0-4.95c1.37-1.37,3.58-1.37,4.95,0c1.12,1.12,1.31,2.8,0.59,4.13l1.88,1.88 L15.04,19.45z M13,9V3.5L18.5,9H13z"/><circle cx="11.5" cy="14.5" r="1.5"/></g></g></svg>'  # noqa



"""The class implementation for the `add_link` material icon.
"""

from apysc._display.fixed_html_svg_icon_base import FixedHtmlSvgIconBase


class MaterialAddLinkIcon(FixedHtmlSvgIconBase):
    """
    The class implementation for the `add_link` material icon.
    """

    def _get_fixed_svg_icon_html(self) -> str:
        """
        Get a fixed SVG icon HTML string.

        Returns
        -------
        fixed_svg_icon_html : str
            Fixed SVG icon HTML string.

        References
        ----------
        - Material icon document
            - https://simon-ritchie.github.io/apysc/en/material_icon.html
        - Material Symbols & Icons, APACHE LICENSE, VERSION 2.0
            - https://fonts.google.com/icons?icon.size=24&icon.color=%23e8eaed
            - https://www.apache.org/licenses/LICENSE-2.0.html
        """
        return '<svg xmlns="http://www.w3.org/2000/svg" height="24" viewBox="0 0 24 24" width="24"><path d="M0 0h24v24H0" fill="none"/><path d="M8 11h8v2H8zm12.1 1H22c0-2.76-2.24-5-5-5h-4v1.9h4c1.71 0 3.1 1.39 3.1 3.1zM3.9 12c0-1.71 1.39-3.1 3.1-3.1h4V7H7c-2.76 0-5 2.24-5 5s2.24 5 5 5h4v-1.9H7c-1.71 0-3.1-1.39-3.1-3.1zM19 12h-2v3h-3v2h3v3h2v-3h3v-2h-3z"/></svg>'  # noqa



"""This module is for the translation mapping data of the
following document:

Document file: graphics_base_scale_from_point.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {
    ##################################################
    "# GraphicsBase get_scale_from_point and set_scale_from_point interfaces": "# GraphicsBase クラスの get_scale_from_point と set_scale_from_point のインターフェイス",  # noqa
    ##################################################
    "This page explains the `GraphicsBase` class (base class of each graphic, such as the `Rectangle`) `get_scale_x_from_point`, `get_scale_y_from_point`, `set_scale_x_from_point`, and `set_scale_y_from_point` method interfaces.": "このページでは`GraphicsBase`クラス（`Rectangle`などのグラフィッククラスの基底クラス）の`get_scale_x_from_point`、`get_scale_y_from_point`、`set_scale_x_from_point`、`set_scale_y_from_point`メソッドの各インターフェイスについて説明します。",  # noqa
    ##################################################
    "## What interfaces are these?": "## 各インターフェイスの概要",
    ##################################################
    "The `set_scale_x_from_point` method changes the object's horizontal scale from a given x-coordinate. Similarly, the `set_scale_y_from_point` method changes the object's vertical scale from a given y-coordinate.": "`set_scale_x_from_point`メソッドは指定された座標を基準にオブジェクトの水平方向の拡縮を変更します。同様に`set_scale_y_from_point`メソッドは指定された座標を基準にオブジェクトの垂直方向の拡縮を変更します。",  # noqa
    ##################################################
    "The `scale_x_from_center` and `scale_y_from_center` interfaces are property, but the `set_scale_x_from_point` and `set_scale_y_from_point` interfaces are methods since these interfaces require a coodinate argument.": "`scale_x_from_center`や`scale_y_from_center`のインターフェイスは属性になっていますが、`set_scale_x_from_point`や`set_scale_y_from_point`のインターフェイスは座標の指定が必要なためメソッドのインターフェイスになっています。",  # noqa
    ##################################################
    "Similarly, the `get_scale_x_from_point` and `get_scale_y_from_point` methods will return the current scale from a given point. These interfaces also require a coordinate argument.": "同様に`get_scale_x_from_point`と`get_scale_y_from_point`のメソッドは指定された座標における拡縮値を返却します。これらのインターフェイスも引数に座標の指定が必要になります。",  # noqa
    ##################################################
    "Return value is set for each coordinate. For example, if you set the scale-x value at the 50px x-coordinate, 100px x-coordinate scale will not be affected.": "返却値は各座標ごとに設定されます。例えば水平方向の拡縮を50pxのX座標の位置で設定した場合、100pxのX座標の位置における拡縮値には影響しません。",  # noqa
    ##################################################
    "## Basic usage": "## 基本的な使い方",
    ##################################################
    "The `get_scale_x_from_point` method requires the `x` argument (`Int` value), and the `set_scale_x_from_point` requires the `scale_x` (`Number` value) and `x` arguments.": "`get_scale_x_from_point`メソッドは`Int`型の`x`引数の指定を必要とし、`set_scale_x_from_point`メソッドは`Number`型の`scale_x`と`x`の各引く数の指定が必要になります。",  # noqa
    ##################################################
    "The following example creates three rectangles and increments (or decrements) for each rectangle scale-x value. The top rectangle scales from the left x position. The middle one scales from the center x. And the bottom one scales from the right x.": "以下のコード例では3つの四角を生成し水平方向の拡縮を増減させています。上の四角は左端を基準に拡縮を、真ん中の四角は中央を基準に拡縮を、そして下の四角では右右端を基準に拡縮を行っています。",  # noqa
    ##################################################
    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass _Options(TypedDict):\n    rectangle: ap.Rectangle\n    x: ap.Int\n    direction: ap.Int\n\n\ndef on_timer(e: ap.TimerEvent, options: _Options) -> None:\n    """\n    The handler that the timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = options["rectangle"]\n    x: ap.Int = options["x"]\n    direction: ap.Int = options["direction"]\n    current_scale_x: ap.Number = rectangle.get_scale_x_from_point(x=x)\n    current_scale_x += direction * 0.03\n    rectangle.set_scale_x_from_point(scale_x=current_scale_x, x=x)\n    with ap.If(current_scale_x >= 2.0):\n        direction *= -1\n    with ap.If(current_scale_x <= 0.0):\n        direction *= -1\n\n\nap.Stage(\n    stage_width=150,\n    stage_height=350,\n    background_color=ap.Color("#333"),\n    stage_elem_id="stage",\n)\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=ap.Color("#0af"))\n\ntop_rect: ap.Rectangle = sprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\nmiddle_rect: ap.Rectangle = sprite.graphics.draw_rect(x=50, y=150, width=50, height=50)\nbottom_rect: ap.Rectangle = sprite.graphics.draw_rect(x=50, y=250, width=50, height=50)\n\ntop_rect_direction: ap.Int = ap.Int(1)\noptions: _Options = {\n    "rectangle": top_rect,\n    "x": ap.Int(50),\n    "direction": top_rect_direction,\n}\ntop_rect_timer: ap.Timer = ap.Timer(on_timer, delay=ap.FPS.FPS_60, options=options)\ntop_rect_timer.start()\n\nmiddle_rect_direction: ap.Int = ap.Int(1)\noptions = {\n    "rectangle": middle_rect,\n    "x": ap.Int(75),\n    "direction": middle_rect_direction,\n}\nmiddle_rect_timer: ap.Timer = ap.Timer(on_timer, delay=ap.FPS.FPS_60, options=options)\nmiddle_rect_timer.start()\n\nbottom_rect_direction: ap.Int = ap.Int(1)\noptions = {\n    "rectangle": bottom_rect,\n    "x": ap.Int(100),\n    "direction": bottom_rect_direction,\n}\nbottom_rect_timer: ap.Timer = ap.Timer(on_timer, delay=ap.FPS.FPS_60, options=options)\nbottom_rect_timer.start()\n\nap.save_overall_html(dest_dir_path="graphics_base_scale_from_point_basic_usage_x/")\n```': '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass _Options(TypedDict):\n    rectangle: ap.Rectangle\n    x: ap.Int\n    direction: ap.Int\n\n\ndef on_timer(e: ap.TimerEvent, options: _Options) -> None:\n    """\n    The handler that the timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = options["rectangle"]\n    x: ap.Int = options["x"]\n    direction: ap.Int = options["direction"]\n    current_scale_x: ap.Number = rectangle.get_scale_x_from_point(x=x)\n    current_scale_x += direction * 0.03\n    rectangle.set_scale_x_from_point(scale_x=current_scale_x, x=x)\n    with ap.If(current_scale_x >= 2.0):\n        direction *= -1\n    with ap.If(current_scale_x <= 0.0):\n        direction *= -1\n\n\nap.Stage(\n    stage_width=150,\n    stage_height=350,\n    background_color=ap.Color("#333"),\n    stage_elem_id="stage",\n)\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=ap.Color("#0af"))\n\ntop_rect: ap.Rectangle = sprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\nmiddle_rect: ap.Rectangle = sprite.graphics.draw_rect(x=50, y=150, width=50, height=50)\nbottom_rect: ap.Rectangle = sprite.graphics.draw_rect(x=50, y=250, width=50, height=50)\n\ntop_rect_direction: ap.Int = ap.Int(1)\noptions: _Options = {\n    "rectangle": top_rect,\n    "x": ap.Int(50),\n    "direction": top_rect_direction,\n}\ntop_rect_timer: ap.Timer = ap.Timer(on_timer, delay=ap.FPS.FPS_60, options=options)\ntop_rect_timer.start()\n\nmiddle_rect_direction: ap.Int = ap.Int(1)\noptions = {\n    "rectangle": middle_rect,\n    "x": ap.Int(75),\n    "direction": middle_rect_direction,\n}\nmiddle_rect_timer: ap.Timer = ap.Timer(on_timer, delay=ap.FPS.FPS_60, options=options)\nmiddle_rect_timer.start()\n\nbottom_rect_direction: ap.Int = ap.Int(1)\noptions = {\n    "rectangle": bottom_rect,\n    "x": ap.Int(100),\n    "direction": bottom_rect_direction,\n}\nbottom_rect_timer: ap.Timer = ap.Timer(on_timer, delay=ap.FPS.FPS_60, options=options)\nbottom_rect_timer.start()\n\nap.save_overall_html(dest_dir_path="graphics_base_scale_from_point_basic_usage_x/")\n```',  # noqa
    ##################################################
    "The `get_scale_y_from_point` and `set_scale_y_from_point` methods have the similar arguments, `scale_y` and `y`. These interfaces work the same way as the x-axis interfaces, except that the axis directions are different.": "似たような形で`get_scale_y_from_point`と`set_scale_y_from_point`のメソッドは`scale_y`と`y`の引数を必要とします。これらは拡縮の方向が垂直方向になっている以外は水平方向のインターフェイスと同じように動作します。",  # noqa
    ##################################################
    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass _Options(TypedDict):\n    rectangle: ap.Rectangle\n    y: ap.Int\n    direction: ap.Int\n\n\ndef on_timer(e: ap.TimerEvent, options: _Options) -> None:\n    """\n    The handler that the timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = options["rectangle"]\n    y: ap.Int = options["y"]\n    direction: ap.Int = options["direction"]\n    current_scale_y: ap.Number = rectangle.get_scale_y_from_point(y=y)\n    current_scale_y += direction * 0.03\n    rectangle.set_scale_y_from_point(scale_y=current_scale_y, y=y)\n    with ap.If(current_scale_y >= 2.0):\n        direction *= -1\n    with ap.If(current_scale_y <= 0.0):\n        direction *= -1\n\n\nap.Stage(\n    stage_width=150,\n    stage_height=150,\n    background_color=ap.Color("#333"),\n    stage_elem_id="stage",\n)\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=ap.Color("#0af"))\n\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\ndirection: ap.Int = ap.Int(1)\noptions: _Options = {"rectangle": rectangle, "y": ap.Int(50), "direction": direction}\ntimer: ap.Timer = ap.Timer(on_timer, delay=ap.FPS.FPS_60, options=options)\ntimer.start()\n\nap.save_overall_html(dest_dir_path="graphics_base_scale_from_point_basic_usage_y/")\n```': '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass _Options(TypedDict):\n    rectangle: ap.Rectangle\n    y: ap.Int\n    direction: ap.Int\n\n\ndef on_timer(e: ap.TimerEvent, options: _Options) -> None:\n    """\n    The handler that the timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = options["rectangle"]\n    y: ap.Int = options["y"]\n    direction: ap.Int = options["direction"]\n    current_scale_y: ap.Number = rectangle.get_scale_y_from_point(y=y)\n    current_scale_y += direction * 0.03\n    rectangle.set_scale_y_from_point(scale_y=current_scale_y, y=y)\n    with ap.If(current_scale_y >= 2.0):\n        direction *= -1\n    with ap.If(current_scale_y <= 0.0):\n        direction *= -1\n\n\nap.Stage(\n    stage_width=150,\n    stage_height=150,\n    background_color=ap.Color("#333"),\n    stage_elem_id="stage",\n)\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=ap.Color("#0af"))\n\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\ndirection: ap.Int = ap.Int(1)\noptions: _Options = {"rectangle": rectangle, "y": ap.Int(50), "direction": direction}\ntimer: ap.Timer = ap.Timer(on_timer, delay=ap.FPS.FPS_60, options=options)\ntimer.start()\n\nap.save_overall_html(dest_dir_path="graphics_base_scale_from_point_basic_usage_y/")\n```',  # noqa
    ##################################################
    "## get_scale_x_from_point API": "## get_scale_x_from_point API",
    ##################################################
    '<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>': '<span class="inconspicuous-txt">特記事項: このAPIドキュメントはドキュメントビルド用のスクリプトによって自動で生成・同期されています。そのためもしかしたらこの節の内容は前節までの内容と重複している場合があります。</span>',  # noqa
    ##################################################
    "**[Interface summary]**": "**[インターフェイス概要]**",
    ##################################################
    "Get a scale-x value from the given x-coordinate.<hr>": "指定されたX座標を基準として水平方向の拡縮の値を取得します。<hr>",  # noqa
    ##################################################
    "**[Parameters]**": "**[引数]**",
    ##################################################
    "- `x`: Number": "- `x`: Number",
    ##################################################
    "  - X-coordinate.": "  - X座標。",
    ##################################################
    "<hr>": "<hr>",
    ##################################################
    "**[Returns]**": "**[返却値]**",
    ##################################################
    "- `scale_x`: Number": "- `scale_x`: Number",
    ##################################################
    "  - Scale-x value from the given x-coordinate.": "  - 指定されたX座標を基準とした水平方向の拡縮値。",
    ##################################################
    "<hr>": "<hr>",
    ##################################################
    "**[Notes]**": "**[特記事項]**",
    ##################################################
    "The scale's minimum value is almost zero, and it does not become negative.<hr>": "拡縮の最小値はほぼ0となり、その値は負の値にはなりません。<hr>",  # noqa
    ##################################################
    "**[Examples]**": "**[コードサンプル]**",
    ##################################################
    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=ap.Color("#0af"))\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50\n... )\n>>> x: ap.Number = ap.Number(100)\n>>> rectangle.set_scale_x_from_point(scale_x=ap.Number(1.5), x=x)\n>>> rectangle.get_scale_x_from_point(x=x)\nNumber(1.5)\n```': '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=ap.Color("#0af"))\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50\n... )\n>>> x: ap.Number = ap.Number(100)\n>>> rectangle.set_scale_x_from_point(scale_x=ap.Number(1.5), x=x)\n>>> rectangle.get_scale_x_from_point(x=x)\nNumber(1.5)\n```',  # noqa
    ##################################################
    "## set_scale_x_from_point API": "## set_scale_x_from_point API",
    ##################################################
    '<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>': '<span class="inconspicuous-txt">特記事項: このAPIドキュメントはドキュメントビルド用のスクリプトによって自動で生成・同期されています。そのためもしかしたらこの節の内容は前節までの内容と重複している場合があります。</span>',  # noqa
    ##################################################
    "**[Interface summary]**": "**[インターフェイス概要]**",
    ##################################################
    "Update a scale-x value from the given x-coordinate.<hr>": "指定されたX座病を基準とした水平方向の拡縮値を更新します。<hr>",  # noqa
    ##################################################
    "**[Parameters]**": "**[引数]**",
    ##################################################
    "- `scale_x`: Number": "- `scale_x`: Number",
    ##################################################
    "  - Scale-x value to set.": "  - 設定する水平方向の拡縮値。",
    ##################################################
    "- `x`: Number": "- `x`: Number",
    ##################################################
    "  - X-coordinate.": "  - X座標。",
    ##################################################
    "<hr>": "<hr>",
    ##################################################
    "**[Notes]**": "**[特記事項]**",
    ##################################################
    "The scale's minimum value is almost zero, and it does not become negative.<hr>": "拡縮の最小値はほぼ0となり、その値は負の値にはなりません。<hr>",  # noqa
    ##################################################
    "**[Examples]**": "**[コードサンプル]**",
    ##################################################
    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=ap.Color("#0af"))\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50\n... )\n>>> x: ap.Number = ap.Number(100)\n>>> rectangle.set_scale_x_from_point(scale_x=ap.Number(1.5), x=x)\n>>> rectangle.get_scale_x_from_point(x=x)\nNumber(1.5)\n```': '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=ap.Color("#0af"))\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50\n... )\n>>> x: ap.Number = ap.Number(100)\n>>> rectangle.set_scale_x_from_point(scale_x=ap.Number(1.5), x=x)\n>>> rectangle.get_scale_x_from_point(x=x)\nNumber(1.5)\n```',  # noqa
    ##################################################
    "## get_scale_y_from_point API": "## get_scale_y_from_point API",
    ##################################################
    '<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>': '<span class="inconspicuous-txt">特記事項: このAPIドキュメントはドキュメントビルド用のスクリプトによって自動で生成・同期されています。そのためもしかしたらこの節の内容は前節までの内容と重複している場合があります。</span>',  # noqa
    ##################################################
    "**[Interface summary]**": "**[インターフェイス概要]**",
    ##################################################
    "Get a scale-y value from the given y-coordinate.<hr>": "指定されたY座標を基準とした垂直方向の拡縮の値を取得します。<hr>",  # noqa
    ##################################################
    "**[Parameters]**": "**[引数]**",
    ##################################################
    "- `y`: Number": "- `y`: Number",
    ##################################################
    "  - Y-coordinate.": "  - Y座標。",
    ##################################################
    "<hr>": "<hr>",
    ##################################################
    "**[Returns]**": "**[返却値]**",
    ##################################################
    "- `scale_y`: ap.Number": "- `scale_y`: ap.Number",
    ##################################################
    "  - Scale-y value from the given y-coordinate.": "  - 指定されたY座標を基準とした垂直方向の拡縮値。",
    ##################################################
    "<hr>": "<hr>",
    ##################################################
    "**[Examples]**": "**[コードサンプル]**",
    ##################################################
    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=ap.Color("#0af"))\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50\n... )\n>>> y: ap.Number = ap.Number(100)\n>>> rectangle.set_scale_y_from_point(scale_y=ap.Number(1.5), y=y)\n>>> rectangle.get_scale_y_from_point(y=y)\nNumber(1.5)\n```': '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=ap.Color("#0af"))\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50\n... )\n>>> y: ap.Number = ap.Number(100)\n>>> rectangle.set_scale_y_from_point(scale_y=ap.Number(1.5), y=y)\n>>> rectangle.get_scale_y_from_point(y=y)\nNumber(1.5)\n```',  # noqa
    ##################################################
    "## set_scale_y_from_point API": "## set_scale_y_from_point API",
    ##################################################
    '<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>': '<span class="inconspicuous-txt">特記事項: このAPIドキュメントはドキュメントビルド用のスクリプトによって自動で生成・同期されています。そのためもしかしたらこの節の内容は前節までの内容と重複している場合があります。</span>',  # noqa
    ##################################################
    "**[Interface summary]**": "**[インターフェイス概要]**",
    ##################################################
    "Update a scale-y value from the given y-coordinate.<hr>": "指定されたY座標を基準とした垂直方向の拡縮値を更新します。<hr>",  # noqa
    ##################################################
    "**[Parameters]**": "**[引数]**",
    ##################################################
    "- `scale_y`: Number": "- `scale_y`: Number",
    ##################################################
    "  - Scale-y value to set.": "  - 設定すの垂直方向の拡縮値。",
    ##################################################
    "- `y`: Number": "- `y`: Number",
    ##################################################
    "  - Y-coordinate.": "  - Y座標。",
    ##################################################
    "<hr>": "<hr>",
    ##################################################
    "**[Examples]**": "**[コードサンプル]**",
    ##################################################
    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=ap.Color("#0af"))\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50\n... )\n>>> y: ap.Number = ap.Number(100)\n>>> rectangle.set_scale_y_from_point(scale_y=ap.Number(1.5), y=y)\n>>> rectangle.get_scale_y_from_point(y=y)\nNumber(1.5)\n```': '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=ap.Color("#0af"))\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50\n... )\n>>> y: ap.Number = ap.Number(100)\n>>> rectangle.set_scale_y_from_point(scale_y=ap.Number(1.5), y=y)\n>>> rectangle.get_scale_y_from_point(y=y)\nNumber(1.5)\n```',  # noqa
}



"""The class implementation for the `MultiLineText` class.
"""

from typing import Optional
from typing import Union

from apysc._color.color import Color
from apysc._color.colors import Colors
from apysc._display.add_foreign_object_child_mixin import AddForeignObjectChildMixIn
from apysc._display.add_to_parent_mixin import AddToParentMixIn
from apysc._display.append_foreign_object_constructor_expression_mixin import (
    AppendForeignObjectConstructorExpressionMixIn,
)
from apysc._display.child_mixin import ChildMixIn
from apysc._display.css_mixin import CssMixIn
from apysc._display.css_text_align import CssTextAlign
from apysc._display.css_text_align_last import CssTextAlignLast
from apysc._display.display_object import DisplayObject
from apysc._display.opacity_css_mixin import OpacityCssMixIn
from apysc._display.set_overflow_visible_setting_mixin import (
    SetOverflowVisibleSettingMixIn,
)
from apysc._display.svg_foreign_object_child_mixin import SvgForeignObjectChildMixIn
from apysc._display.svg_foreign_object_initialize_width_mixin import (
    SvgForeignObjectInitializeWidthMixIn,
)
from apysc._display.svg_foreign_object_text_mixin import SvgForeignObjectTextMixIn
from apysc._display.text_align_css_mixin import TextAlignCssMixIn
from apysc._display.text_align_last_css_mixin import TextAlignLastCssMixIn
from apysc._display.text_bold_css_mixin import TextBoldCssMixIn
from apysc._display.text_decoration_underline_css_mixin import (
    TextDecorationUnderlineCssMixIn,
)
from apysc._display.text_fill_color_css_mixin import TextFillColorCssMixIn
from apysc._display.text_font_size_css_mixin import TextFontSizeCssMixIn
from apysc._display.text_italic_css_mixin import TextItalicCssMixIn
from apysc._display.text_line_height_css_mixin import TextLineHeightCssMixIn
from apysc._display.width_mixin import WidthMixIn
from apysc._display.x_mixin import XMixIn
from apysc._display.y_mixin import YMixIn
from apysc._loop.initialize_with_base_value_interface import (
    InitializeWithBaseValueInterface,
)
from apysc._type.boolean import Boolean
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.string import String
from apysc._type.variable_name_suffix_mixin import VariableNameSuffixMixIn
from apysc._validation import arg_validation_decos


class MultiLineText(
    XMixIn,
    YMixIn,
    DisplayObject,
    WidthMixIn,
    VariableNameSuffixMixIn,
    CssMixIn,
    SetOverflowVisibleSettingMixIn,
    InitializeWithBaseValueInterface,
    AddForeignObjectChildMixIn,
    AppendForeignObjectConstructorExpressionMixIn,
    SvgForeignObjectInitializeWidthMixIn,
    SvgForeignObjectTextMixIn,
    SvgForeignObjectChildMixIn,
    AddToParentMixIn,
    TextFillColorCssMixIn,
    OpacityCssMixIn,
    TextBoldCssMixIn,
    TextItalicCssMixIn,
    TextAlignCssMixIn,
    TextAlignLastCssMixIn,
    TextDecorationUnderlineCssMixIn,
    TextFontSizeCssMixIn,
    TextLineHeightCssMixIn,
):
    """
    The class implementation for a multiline text element.

    References
    ----------
    - MultiLineText class
        - https://simon-ritchie.github.io/apysc/en/multi_line_text.html
    - Text fill_color property
        - https://simon-ritchie.github.io/apysc/en/text_fill_color.html
    - Text fill_alpha property
        - https://simon-ritchie.github.io/apysc/en/text_fill_alpha.html
    - Text bold property
        - https://simon-ritchie.github.io/apysc/en/text_bold.html
    - Text italic property
        - https://simon-ritchie.github.io/apysc/en/text_italic.html
    - text_align property
        - https://simon-ritchie.github.io/apysc/en/text_align.html
    - text_align_last property
        - https://simon-ritchie.github.io/apysc/en/text_align_last.html
    - Text font_size property
        - https://simon-ritchie.github.io/apysc/en/text_font_size.html
    - Text line_height property
        - https://simon-ritchie.github.io/apysc/en/text_line_height.html
    """

    # text
    @arg_validation_decos.is_string(arg_position_index=1, optional=False)
    # x
    @arg_validation_decos.is_num(arg_position_index=2, optional=False)
    # y
    @arg_validation_decos.is_num(arg_position_index=3, optional=False)
    # width
    @arg_validation_decos.is_integer(arg_position_index=4, optional=False)
    # font_size
    @arg_validation_decos.is_integer(arg_position_index=5, optional=False)
    # fill_color
    @arg_validation_decos.is_color(arg_position_index=6, optional=False)
    # fill_alpha
    @arg_validation_decos.is_num(arg_position_index=7, optional=False)
    # bold
    @arg_validation_decos.is_boolean(arg_position_index=8, optional=False)
    # italic
    @arg_validation_decos.is_boolean(arg_position_index=9, optional=False)
    # text_align
    @arg_validation_decos.is_css_text_align(arg_position_index=10)
    # text_align_last
    @arg_validation_decos.is_css_text_align_last(arg_position_index=11)
    # underline
    @arg_validation_decos.is_boolean(arg_position_index=12, optional=False)
    # line_height
    @arg_validation_decos.is_num(arg_position_index=13, optional=False)
    # parent
    @arg_validation_decos.is_display_object_container(
        arg_position_index=14, optional=True
    )
    def __init__(
        self,
        *,
        text: Union[str, String],
        x: Union[float, Number] = 0,
        y: Union[float, Number] = 0,
        width: Union[int, Int] = 200,
        font_size: Union[int, Int] = 16,
        fill_color: Color = Colors.GRAY_666666,
        fill_alpha: Union[float, Number] = 1.0,
        bold: Union[bool, Boolean] = False,
        italic: Union[bool, Boolean] = False,
        text_align: CssTextAlign = CssTextAlign.LEFT,
        text_align_last: CssTextAlignLast = CssTextAlignLast.AUTO,
        underline: Union[bool, Boolean] = False,
        line_height: Union[float, Number] = 1.5,
        parent: Optional[ChildMixIn] = None,
        variable_name_suffix: str = "",
    ) -> None:
        """
        The class implementation for a multiline text element.

        Parameters
        ----------
        text : Union[str, String]
            Text to display. An HTML tag is available.
        x : Union[float, Number], default 0
            X-coordinate.
        y : Union[float, Number], default 0
            Y-coordinate.
        width : Union[int, Int], default 200
            Width of the text to wrap.
        font_size : Union[int, Int], default 16
            Font size.
        fill_color : Color, default Colors.GRAY_666666
            Text color.
        fill_alpha : Union[float, Number], default 1.0
            Text alpha (opacity). The minimum value is 0.0 (transparent),
            and the maximum value is 1.0 (solid).
        bold : Union[bool, Boolean], default False
            Whether to display the text in bold.
        italic : Union[bool, Boolean], default False
            Whether to display the text in italic.
        text_align : CssTextAlign, default `CssTextAlign.LEFT`
            Text align setting.
        text_align_last : CssTextAlignLast, default `CssTextAlignLast.AUTO`
            Last line's text-align setting.
        underline : Union[bool, Boolean], default False
            Whether to display the text's underline.
        line_height : Union[float, Number], default 1.5
            A line-height (text-leading) setting.
        parent : ChildMixIn or None, default None
            A parent instance to add this instance.
            If the specified value is None, this interface uses
            a stage instance.
        variable_name_suffix : str, default ""
            A JavaScript variable name suffix string.
            This setting is sometimes useful for JavaScript debugging.

        References
        ----------
        - MultiLineText class
            - https://simon-ritchie.github.io/apysc/en/multi_line_text.html
        - Text fill_color property
            - https://simon-ritchie.github.io/apysc/en/text_fill_color.html
        - Text fill_alpha property
            - https://simon-ritchie.github.io/apysc/en/text_fill_alpha.html
        - Text bold property
            - https://simon-ritchie.github.io/apysc/en/text_bold.html
        - Text italic property
            - https://simon-ritchie.github.io/apysc/en/text_italic.html
        - text_align property
            - https://simon-ritchie.github.io/apysc/en/text_align.html
        - text_align_last property
            - https://simon-ritchie.github.io/apysc/en/text_align_last.html
        - Text font_size property
            - https://simon-ritchie.github.io/apysc/en/text_font_size.html
        - Text line_height property
            - https://simon-ritchie.github.io/apysc/en/text_line_height.html

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage(
        ...     background_color=ap.Color("#333"),
        ...     stage_width=300,
        ...     stage_height=100,
        ...     stage_elem_id="stage",
        ... )
        >>> multi_line_text: ap.MultiLineText = ap.MultiLineText(
        ...     text="Lorem ipsum dolor sit amet, consectetur adipiscing elit, "
        ...     "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. "
        ...     "Ut enim ad minim veniam",
        ...     width=300,
        ...     font_size=16,
        ...     fill_color=ap.Color("#00aaff"),
        ...     x=20,
        ...     y=20,
        ... )
        >>> multi_line_text.fill_color
        Color("#00aaff")
        """
        from apysc._converter import to_apysc_val_from_builtin
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names

        self._variable_name_suffix = variable_name_suffix
        variable_name: str = expression_variables_util.get_next_variable_name(
            type_name=var_names.MULTILINE_TEXT,
        )
        super(MultiLineText, self).__init__(variable_name=variable_name)
        self._initialize_width(width=width)
        self._append_foreign_object_constructor_expression()
        self._add_to_parent(parent=parent)
        self._set_overflow_visible_setting()
        self._initialize_text(text=text)
        self._initialize_svg_foreign_object_child(
            html_str=self._text, variable_name_suffix=variable_name_suffix
        )
        self.font_size = to_apysc_val_from_builtin.get_copied_int_from_builtin_val(
            integer=font_size
        )
        self.x = to_apysc_val_from_builtin.get_copied_number_from_builtin_val(
            float_or_num=x
        )
        self.y = to_apysc_val_from_builtin.get_copied_number_from_builtin_val(
            float_or_num=y
        )
        self.fill_color = fill_color.copy()
        self.fill_alpha = to_apysc_val_from_builtin.get_copied_number_from_builtin_val(
            float_or_num=fill_alpha,
            variable_name_suffix=variable_name_suffix,
        )
        self.bold = to_apysc_val_from_builtin.get_copied_boolean_from_builtin_val(
            bool_val=bold,
            variable_name_suffix=variable_name_suffix,
        )
        self.italic = to_apysc_val_from_builtin.get_copied_boolean_from_builtin_val(
            bool_val=italic,
            variable_name_suffix=variable_name_suffix,
        )
        self.text_align = text_align
        self.text_align_last = text_align_last
        self.underline = to_apysc_val_from_builtin.get_copied_boolean_from_builtin_val(
            bool_val=underline,
            variable_name_suffix=variable_name_suffix,
        )
        self.line_height = to_apysc_val_from_builtin.get_copied_number_from_builtin_val(
            float_or_num=line_height,
            variable_name_suffix=variable_name_suffix,
        )

    @classmethod
    def _initialize_with_base_value(cls) -> "MultiLineText":
        """
        Initialize this class with a base value(s).

        Returns
        -------
        text : MultiLineText
            An initialized instance.
        """
        text: MultiLineText = cls(text="", width=0)
        return text



"""Class implementations for the year-related mix-in.
"""

from typing import Union

from typing_extensions import final

from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.int import Int
from apysc._type.revert_mixin import RevertMixIn
from apysc._type.variable_name_mixin import VariableNameMixIn
from apysc._type.variable_name_suffix_attr_or_var_mixin import (
    VariableNameSuffixAttrOrVarMixIn,
)
from apysc._validation import arg_validation_decos


class YearMixIn(VariableNameMixIn, VariableNameSuffixAttrOrVarMixIn, RevertMixIn):
    _initial_year: Union[int, Int]
    _year: Int

    @final
    @add_debug_info_setting(module_name=__name__)
    @arg_validation_decos.is_four_digit_year(arg_position_index=1)
    def _set_init_year_value(self, *, year: Union[int, Int]) -> None:
        """
        Set an initial year value.

        Parameters
        ----------
        year : Union[int, Int]
            A year value to set.
        """
        from apysc._converter.to_apysc_val_from_builtin import (
            get_copied_int_from_builtin_val,
        )

        self._initial_year = year
        suffix: str = self._get_attr_or_variable_name_suffix(value_identifier="year")
        self._year = get_copied_int_from_builtin_val(
            integer=year, variable_name_suffix=suffix
        )

    @final
    @add_debug_info_setting(module_name=__name__)
    def _get_init_year_argument_expression(self) -> str:
        """
        Get an initial year's argument expression string.

        Returns
        -------
        expression : str
            A created expression string.
        """
        if isinstance(self._initial_year, Int):
            return f"{self._initial_year.variable_name}"
        return f"{self._year._value}"

    @property
    @add_debug_info_setting(module_name=__name__)
    def year(self) -> Int:
        """
        Get a current year's value.

        Returns
        -------
        year : Int
            A current year value.

        References
        ----------
        - DateTime class year property
            - https://simon-ritchie.github.io/apysc/en/datetime_year.html

        Examples
        --------
        >>> import apysc as ap
        >>> _ = ap.Stage()
        >>> datetime_: ap.DateTime = ap.DateTime(year=2022, month=12, day=1)
        >>> datetime_.year
        Int(2022)
        >>> datetime_.year = ap.Int(2023)
        >>> datetime_.year
        Int(2023)
        """
        copied_year_val: Int = self._year.copy()
        self._append_year_getter_expression(year_val=copied_year_val)
        return copied_year_val

    @year.setter
    @arg_validation_decos.is_four_digit_year(arg_position_index=1)
    @add_debug_info_setting(module_name=__name__)
    def year(self, value: Int) -> None:
        """
        Set a year value.

        Parameters
        ----------
        value : Int
            A year value to set.

        References
        ----------
        - DateTime class year property
            - https://simon-ritchie.github.io/apysc/en/datetime_year.html

        Examples
        --------
        >>> import apysc as ap
        >>> _ = ap.Stage()
        >>> datetime_: ap.DateTime = ap.DateTime(year=2022, month=12, day=1)
        >>> datetime_.year
        Int(2022)
        >>> datetime_.year = ap.Int(2023)
        >>> datetime_.year
        Int(2023)
        """
        self._year = value.copy()
        self._append_year_setter_expression(year_val=value)

    @final
    @arg_validation_decos.is_four_digit_year(arg_position_index=1)
    @add_debug_info_setting(module_name=__name__)
    def _append_year_getter_expression(self, *, year_val: Int) -> None:
        """
        Append a year's getter expression string.

        Parameters
        ----------
        year_val : Int
            A year value to use in an expression.
        """
        from apysc._expression import expression_data_util

        expression: str = (
            f"{year_val.variable_name} = {self.variable_name}.getFullYear();"
        )
        expression_data_util.append_js_expression(expression=expression)

    @final
    @arg_validation_decos.is_four_digit_year(arg_position_index=1)
    @add_debug_info_setting(module_name=__name__)
    def _append_year_setter_expression(self, *, year_val: Int) -> None:
        """
        Append a year's setter expression string.

        Parameters
        ----------
        year_val : Int
            A year value to use in an expression.
        """
        from apysc._expression import expression_data_util

        expression: str = f"{self.variable_name}.setFullYear({year_val.variable_name});"
        expression_data_util.append_js_expression(expression=expression)

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._year._run_all_make_snapshot_methods(snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if a snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._year._run_all_revert_methods(snapshot_name=snapshot_name)



"""The class implementation for the outlined `arrow_circle_down` material icon.
"""

from apysc._display.fixed_html_svg_icon_base import FixedHtmlSvgIconBase


class MaterialArrowCircleDownOutlinedIcon(FixedHtmlSvgIconBase):
    """
    The class implementation for the outlined `arrow_circle_down` material icon.
    """

    def _get_fixed_svg_icon_html(self) -> str:
        """
        Get a fixed SVG icon HTML string.

        Returns
        -------
        fixed_svg_icon_html : str
            Fixed SVG icon HTML string.

        References
        ----------
        - Material icon document
            - https://simon-ritchie.github.io/apysc/en/material_icon.html
        - Material Symbols & Icons, APACHE LICENSE, VERSION 2.0
            - https://fonts.google.com/icons?icon.size=24&icon.color=%23e8eaed
            - https://www.apache.org/licenses/LICENSE-2.0.html
        """
        return '<svg xmlns="http://www.w3.org/2000/svg" enable-background="new 0 0 24 24" height="24" viewBox="0 0 24 24" width="24"><g><rect fill="none" height="24" width="24"/><path d="M12,4c4.41,0,8,3.59,8,8s-3.59,8-8,8s-8-3.59-8-8S7.59,4,12,4 M12,2C6.48,2,2,6.48,2,12c0,5.52,4.48,10,10,10 c5.52,0,10-4.48,10-10C22,6.48,17.52,2,12,2L12,2z M13,12l0-4h-2l0,4H8l4,4l4-4H13z"/></g></svg>'  # noqa



"""This module is for the translation mapping data of the
following document:

Document file: trace.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {
    ##################################################
    "# trace interface": "# trace インターフェイス",
    ##################################################
    "This page explains the `trace` function interface.": "このページでは`trace`関数のインターフェイスについて説明します。",  # noqa
    ##################################################
    "## What interface is this?": "## インターフェイス概要",
    ##################################################
    "The `trace` function interface displays any message on the browser console. This interface behaves like the JavaScript `console.log` function.": "`trace`関数のインターフェイスは任意のメッセージをブラウザのコンソール上に表示します。このインターフェイスはJavaScriptの`console.log`の関数と同じような挙動をします。",  # noqa
    ##################################################
    "Also, this interface displays Python's file name, caller information, and line number.": "また、このインターフェイスはPythonのファイル名、呼び出し元の情報、行番号などの情報も表示します。",  # noqa
    ##################################################
    "## Note for the print alias": "## print関数のエイリアスに対する特記事項",
    ##################################################
    "The `ap.print` function is the alias of the `trace` function. Therefore, It behaves the same as the `trace` function.": "`ap.print`関数は`trace`関数のエイリアスとなります。そのため、その関数は`trace`関数と同じように動作します。",  # noqa
    ##################################################
    "## Basic usage": "## 基本的な使い方",
    ##################################################
    "The `trace` function can accept any number of arguments and various value types.": "`trace`関数は任意の数の引数を受け付け、そして様々な型の値を指定することができます。",  # noqa
    ##################################################
    "The following example draws the rectangle. Then, the handler displays the message on the browser console when you click it(please press the F12 key).": "伊賀のコード例では四角を描画し、四角をクリックした際にブラウザのコンソール上にメッセージを表示するようにしています（F12キーを押してコンソールを開いてご確認ください）。",  # noqa
    ##################################################
    '```py\n# runnable\nimport apysc as ap\n\n\ndef on_rectangle_click(e: ap.MouseEvent[ap.Rectangle], options: dict) -> None:\n    """\n    The handler that the rectangle calls when clicked.\n\n    Parameters\n    ----------\n    e : MouseEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    ap.trace("Hello apysc!", "Rectangle width:", e.this.width)\n\n\nap.Stage(\n    stage_width=150,\n    stage_height=150,\n    background_color=ap.Color("#333"),\n    stage_elem_id="stage",\n)\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=ap.Color("#0af"))\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\nrectangle.click(on_rectangle_click)\n\nap.save_overall_html(dest_dir_path="trace_basic_usage/")\n```': '```py\n# runnable\nimport apysc as ap\n\n\ndef on_rectangle_click(e: ap.MouseEvent[ap.Rectangle], options: dict) -> None:\n    """\n    The handler that the rectangle calls when clicked.\n\n    Parameters\n    ----------\n    e : MouseEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    ap.trace("Hello apysc!", "Rectangle width:", e.this.width)\n\n\nap.Stage(\n    stage_width=150,\n    stage_height=150,\n    background_color=ap.Color("#333"),\n    stage_elem_id="stage",\n)\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=ap.Color("#0af"))\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\nrectangle.click(on_rectangle_click)\n\nap.save_overall_html(dest_dir_path="trace_basic_usage/")\n```',  # noqa
    ##################################################
    "## trace API": "## trace API",
    ##################################################
    '<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>': '<span class="inconspicuous-txt">特記事項: このAPIドキュメントはドキュメントビルド用のスクリプトによって自動で生成・同期されています。そのためもしかしたらこの節の内容は前節までの内容と重複している場合があります。</span>',  # noqa
    ##################################################
    "**[Interface summary]**": "**[インターフェイス概要]**",
    ##################################################
    "Display arguments information to console. This function saves a JavaScript `console.log` expression.<hr>": "引数に指定された値の情報をコンソールへ表示します。この関数はJavaScriptの`console.log`に該当するコードを保存します。<hr>",  # noqa
    ##################################################
    "**[Parameters]**": "**[引数]**",
    ##################################################
    "- `*args`: list": "- `*args`: list",
    ##################################################
    "  - Any arguments to display to console.": "  - コンソール上に表示する任意の引数の値。",
    ##################################################
    "<hr>": "<hr>",
    ##################################################
    "**[Notes]**": "**[特記事項]**",
    ##################################################
    "The `ap.print` function is the alias of the `trace` function.<hr>": "`ap.print`関数は`trace`関数のエイリアスとなります。<hr>",  # noqa
    ##################################################
    "**[Examples]**": "**[コードサンプル]**",
    ##################################################
    '```py\n>>> import apysc as ap\n>>> _ = ap.Stage()\n>>> int_val: ap.Int = ap.Int(10)\n>>> ap.trace("Int value is:", int_val)\n```': '```py\n>>> import apysc as ap\n>>> _ = ap.Stage()\n>>> int_val: ap.Int = ap.Int(10)\n>>> ap.trace("Int value is:", int_val)\n```',  # noqa
}



"""The class implementation for the outlined `visibility` material icon.
"""

from apysc._display.fixed_html_svg_icon_base import FixedHtmlSvgIconBase


class MaterialVisibilityOutlinedIcon(FixedHtmlSvgIconBase):
    """
    The class implementation for the outlined `visibility` material icon.
    """

    def _get_fixed_svg_icon_html(self) -> str:
        """
        Get a fixed SVG icon HTML string.

        Returns
        -------
        fixed_svg_icon_html : str
            Fixed SVG icon HTML string.

        References
        ----------
        - Material icon document
            - https://simon-ritchie.github.io/apysc/en/material_icon.html
        - Material Symbols & Icons, APACHE LICENSE, VERSION 2.0
            - https://fonts.google.com/icons?icon.size=24&icon.color=%23e8eaed
            - https://www.apache.org/licenses/LICENSE-2.0.html
        """
        return '<svg xmlns="http://www.w3.org/2000/svg" height="24" viewBox="0 0 24 24" width="24"><path d="M0 0h24v24H0V0z" fill="none"/><path d="M12 6c3.79 0 7.17 2.13 8.82 5.5C19.17 14.87 15.79 17 12 17s-7.17-2.13-8.82-5.5C4.83 8.13 8.21 6 12 6m0-2C7 4 2.73 7.11 1 11.5 2.73 15.89 7 19 12 19s9.27-3.11 11-7.5C21.27 7.11 17 4 12 4zm0 5c1.38 0 2.5 1.12 2.5 2.5S13.38 14 12 14s-2.5-1.12-2.5-2.5S10.62 9 12 9m0-2c-2.48 0-4.5 2.02-4.5 4.5S9.52 16 12 16s4.5-2.02 4.5-4.5S14.48 7 12 7z"/></svg>'  # noqa



"""This module is for the class implementation of
the `DeletedObjectMixIn`.
"""

import inspect
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from typing_extensions import final

from apysc._type.revert_mixin import RevertMixIn


class _DisabledObjectError(Exception):
    pass


class DeletedObjectMixIn(RevertMixIn):
    __is_deleted_object: bool = False

    @property
    def _is_deleted_object(self) -> bool:
        """
        Get a boolean indicating whether this object is
        deleted object or not.

        Returns
        -------
        _is_deleted_object : bool
            A boolean indicating whether this object is
            deleted object or not.
        """
        return self.__is_deleted_object

    @_is_deleted_object.setter
    def _is_deleted_object(self, value: bool) -> None:
        """
        Set a boolean indicating whether this object is
        deleted object or not.

        Parameters
        ----------
        value : bool
            A target value to set.
        """
        self.__is_deleted_object = value
        if value:
            self._disable_each_method()

    @final
    def _disabled_method(self, *args: Any, **kwargs: Any) -> None:
        """
        The method to replace each method when this object
        becomes deleted object.

        Raises
        ------
        _DisabledObjectError
            This interface always raises an exception.
        """
        raise _DisabledObjectError(
            "This object has been deleted and cannot manipulate."
        )

    @final
    def _disable_each_method(self) -> None:
        """
        Disable each method of this instance.

        Notes
        -----
        This method disables only public methods (this method skips
        methods which starts with the single underscore character).
        """
        method_members: List[Tuple[str, Callable]] = inspect.getmembers(
            self, predicate=inspect.ismethod
        )
        for method_name, _ in method_members:
            if method_name.startswith("_") and not method_name.startswith("__"):
                continue
            setattr(self, method_name, self._disabled_method)

    _is_deleted_object_snapshot: Optional[Dict[str, bool]] = None

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._set_single_snapshot_val_to_dict(
            dict_name="_is_deleted_object_snapshot",
            value=self.__is_deleted_object,
            snapshot_name=snapshot_name,
        )

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if a snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self.__is_deleted_object = self._get_snapshot_val_if_exists(
            current_value=self.__is_deleted_object,
            snapshot_dict=self._is_deleted_object_snapshot,
            snapshot_name=snapshot_name,
        )
